/*
This program is PC program written to run under LINUX on a terminal.
   Compile with:  cc send_hex.c -o sendhex
   To run: open terminal and run: ./sendhex

   sendhex is a .hex file uploader for the PIC18F47K40 board with bootloader "PIC18F47K40_64MHz_boot.hex" installed.
   Sendhex will ask for .hex file name and then wait for the PIC18F47K40 board to be reset. 
   Program files need to begin at address 0x00004 to use the bootloader, and can extend to 0x1FEFF = 130811 bytes 
   long. The bootoader occupies the upper 256 byte of flash ( two 128 byte blocks).
------------------------------------------------------------------------------------------------------------------
   The Serial port programming function - "set_interface_attribs (int fd, int speed, int parity)":
   Downloaded From: http://stackoverflow.com/questions/6947413/how-to-open-read-and-write-from-serial-port-in-c
------------------------------------------------------------------------------------------------------------------
*/
#define serial_port "/dev/ttyACM0"

#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>

                // global variables
unsigned char hex;          // converted byte from .hex file
char          type = 1;     // .hex file record type
char          position = 0; // current position in .hex file line
char          count = 0;    // length of line in .hex file remaining
char          crc_err = 0;  // flag to show a crc error occured when it = 'e'
unsigned int  addr;         // address from .hex file of current .hex data
char          rcvchr;       // character to be received back from bootloader
unsigned char sndchr;       // byte to be sent to bootloader
unsigned char crc;          // crc of block of data being sent to bootloader
unsigned char check_crc;    // check crc of ,hex line crc
unsigned char buffer[130];  // buffer for 128 byte block of hex characters
unsigned int  buffsize = 0; // number of bytes in buffer[]
unsigned char addressU = 0; // extended address of block: 00-01 = flash, 20 = user ID, 31 = EEPROM, 30 = Config
unsigned int  start_address = 0x0000; // saved 16 bit start of block address
unsigned char start_addressU = 0x00;  // saved 8 bit start of block extended addressU
char filename[80];

FILE *fptr;     // pointer to .hex file "filename"
int   fd;       // serial port /dev/ttyACM0 file descriptor
                // function prototypes
int set_interface_attribs (int fd, int speed, int parity);
void program();
void send_preamble();
void send_block();
char convert_hex(char ch);
void send_byte();
void read_byte(char match);
void print_terminal(long cntr);



int set_interface_attribs (int fd, int speed, int parity)
{
     struct termios tty;
     memset (&tty, 0, sizeof tty);
     if (tcgetattr (fd, &tty) != 0)
        {
        printf("error %d from tcgetattr", errno);
        return -1;
        }
     cfsetospeed (&tty, speed);         // set output to 115200 baud
     cfsetispeed (&tty, speed);         // set input to 115200 baud
     tty.c_cflag = (tty.c_cflag & ~CSIZE) | CS8;  // 8-bit chars
     tty.c_iflag &= ~IGNBRK;            // disable break processing
     tty.c_lflag = 0;                   // no signaling chars, no echo,
                                        // no canonical processing
     tty.c_oflag = 0;                   // no remapping, no delays
     tty.c_cc[VMIN]  = 0;               // read doesn't block
     tty.c_cc[VTIME] = 5;               // 0.5 seconds read timeout
     tty.c_iflag &= ~(IXON | IXOFF | IXANY); // shut off xon/xoff ctrl
     tty.c_cflag |= (CLOCAL | CREAD);   // ignore modem controls,
                                        // enable reading
     tty.c_cflag &= ~(PARENB | PARODD); // shut off parity
     tty.c_cflag |= parity;
     tty.c_cflag &= ~CSTOPB;
     tty.c_cflag &= ~CRTSCTS;
     if (tcsetattr (fd, TCSANOW, &tty) != 0)
        {
        printf("error %d from tcgetattr", errno);
        return -1;
        }
     return 0;
}


void main()
{
     // open and configure serial comm port --------------------------------------------

   fd = open(serial_port, O_RDWR | O_NOCTTY | O_SYNC); // the serial port opened command
   if (fd < 0)
      {
      printf("\n\r!! error %d /dev/ttyACM0 not available !!\n\r", errno);
      return;           // program will exit
      }
   set_interface_attribs (fd, B115200, 0); // 115200 baud no parity
   fcntl(fd, F_SETFL, 0);       // causes read to block until new characters are present
   fcntl(fd, F_SETFL, FNDELAY); // remove this call as it causes read to return immediately

     // open .hex file -----------------------------------------------------------------

   puts("Enter .hex file name");  // enter the .hex file name to upload
   scanf("%s",filename);

   fptr = fopen(filename, "r"); // open text file in read mode
   if(fptr == NULL)
      {
      printf("\n\r!! can't open file %s !!\n\r", filename);
      close(fd);
      return;            // program will exit if fopen fails
      }
   puts("Reset the PIC18F47K40"); // wait forever for PIC18F47K40 to be reset

   read_byte('R');       // read character 'R' sent from bootloader at reset

   sndchr = 0xc1;
   write(fd,&sndchr,1);  // send 0xC1 to bootloader
   usleep(100);          // wait 100 uS for char to transmit

   read_byte('K');       // wait for 0x4B - "OK" from bootloader ('K')

     // program: send .hex file to bootloader ------------------------------------------

   program();            // program() reads .hex file, converts it to hex, uploads it to bootloader

   printf("\n\r---end of file---\n\r");

   close(fd);
   fclose(fptr);
}  // end of main()

       /* There are 4 types of data in inhex32 file for PIC18F47K40.
          They are programmed differently - addressU indicates which type:
            00-01 = program     00 0000 - 01 ffff - programmed in 128 byte blocks
            20    = user ID     20 0000 - 20 000F - programmed by byte
            30    = config      30 0000 - 30 000B - programmed by byte
            31    = eeprom      31 0000 - 31 03ff - programmed by byte */
           
void program()
{
  char ch = 0;            // data byte read from .hex file
  long cntr = 0;          // cntr is used to format data echoed to terminal into 16 byte lines
                          // (cntr is total number of program data bytes sent)
    buffer[0] = 0x93;     // (or 81) code in .hex file being uploaded is required to have been compiled with
    buffer[1] = 0xEF;     // code offset 0x0004, or ORG 0x0004, Otherwise 1st 4 bytes will be overwitten.
    buffer[2] = 0xFF;     // address 0-3 = 93 EF FF F0
    buffer[3] = 0xF0; 

    position = 0;                 // the position within a line in the .hex file
    buffsize = 4;                 // preserve first 4 bytes in first block of data for goto bootloader

    while(feof(fptr) == 0)        // keep reading until end of .hex file
       {
       ch = fgetc(fptr);          // read a character from .hex file
       if( convert_hex(ch) == 1)  // if convert_hex returns a 1 (type = 0) it is data
          {
          if(crc_err == 'e')      // if a crc error was encountered reading .hex file - exit sendhex
             {
             printf("\n\r-crc_error!- EXIT");
             break;               // a break here will close files and exit the program
             }
          if(addressU < 2 )       // if this is Program Flash Memory data. (addresU = 00-01)
             {
             print_terminal(cntr);// print data about to be put in buffer to terminal             
             if(cntr == 0 && addr < 4) // check if first address in the .hex file is less than 4
                {                 // if so it would be overwriten by "goto bootloader"
                printf("\n\r!! %s File's code offset is less than 4 - EXIT !!\n\r", filename);
                break;            // a break here will close files and exit the program
                }
             cntr++;              // update print data line length (count total program data bytes)
             if(addressU == 1 && addr > 0xFEFF) // check if program data will overwite bootloader
                {                 // bootloader is at top of program flash @ address 0x1FEFF
                printf("\n\r!! EXIT - Code address greater than 1FEFF in %s\n\r", filename);
                break;            // a break here will close files and exit the program
                } 
             buffer[(addr & 0x0007f)] = hex; // put data in buffer-limit address range 0 to 127(7f)
             if(buffsize == 0)    // set the start address for the flash block (skipped for first block)
                {                 // in first block, start_addressU and start_address are already 0
                start_addressU = addressU;
                start_address = addr; // the address from .hex file of first byte of data in block
                }                 // in first block goto bootloader instruction is at 0, (it's not in the hex file)
             addr++;              // increment to next .hex file and buffer location
             if((addr & 0x007f) == 0)  // if lower 7 bits of address rolled over to 0 - buffer is full           
                send_block();     // send the 128 byte block to the bootloader
             else
                buffsize = (addr & 0x0007f);  // update the buffer count
             }    // end of if(addressU < 2)
          else    // else, addressU is => than 2 (EEPROM, Config, User ID, Revision ID, Device ID)
             {
             if(buffsize > 0)     // if > 0, there is program data in the buffer (less than a fullblock)
                {                                
                buffsize--;       // decrement the last address update
                send_block();     // send the partial block
                printf("End of Program\n\r");
                }
             if(addressU < 0x3F)  // 30h, 31h, and 20h EEPROM, Config, UserID (3F = revID & devID)
                {               
                send_byte();      // send a single byte to EEPROM or Config
                addr++;           // increment addr to address of next byte in .hex file
                } // end of addressU < 0x3F
             }    // end of else
          }       // end if convert_hex = 1
       }          // end while feop(fptr) == 0
}                 // end program()


char convert_hex(char ch) // decodes inhex32 file and converts data from ascii to hexadecimal
{
  char ret = 0;
      
      position++;         // go to next position of ihx format

      // ------------------------------START OF LINE --------------------------------------

      if(ch == ':')       // start of ihx line identifier
         {
         position = 0;    // set position to position 0 = ihx line length(left)
         check_crc = 0;   // zero check_crc to start a new check_crc for the line 
         }
      //--------------------------CONVERT ascii to hexadecimal ----------------------------

      if(ch > 0x40)       // A is 0x41, a is 0x61
         ch = ch + 9;     // add 9 to convert ch from ascii to hexadecimal
      ch = (0x0f & ch);   // strip the left digit - ch is now converted to hex
      
      // --------------------------------- CRC --------------------------------------------
      
      if(position > 0)               // the line crc is also added here - subtract later
         if((position & 0x01) == 1)  // left nibble     
            check_crc += (ch << 4);  // add to check_crc            
         else                        // right hibble
            check_crc += ch;         // add to check_crc

      // -------------------------------LINE LENGTH ---------------------------------------

      if(position == 1)      // ihx line length(left)
         count = (ch << 4);  // count = the number code bytes in the ihx line
      if(position == 2)      // ihx line length(right)
         count |= ch;            
      // ---------------------------------ADDRESS------------------------------------------

      if(position > 2 && position < 7) // positions 3,4,5,6
         {
         addr = addr << 4;   // shift previous hex nibble over 4 places
         addr |= (unsigned int)(ch & 0x000f);
         }
      // ----------------------------------TYPE -------------------------------------------

      if(position == 7)      // decode type of line (left digit)
         type = (ch << 4);   // (00,01,04) are possibile types

      if(position == 8)      // decode type of line (right digit)
         type |= ch;         // 00 is a data byte, 04 is addressU (2 bytes), 01 is end of file

      //-----------------------------------data--------------------------------------------

      if(position > 8)          // convert 2, 4 bit nibbles to 8 bit hex code
         {
         if( (position & 0x01) == 0x01) // check even/odd position
            hex = (ch << 4);    // odd = left nibble bits 4-7
         else                   // even = low nibble bits 0-3
            if((count--) > 0)   // if this is still code in the ihx line
               {
               if(type == 0)
                  {
                  hex = (hex | ch); // convert the 2 ascii digits to hex value
                  ret = 1;
                  }
               if(type == 4)        // leastsignificant byte is needed (xx00 - xx30)                       
                  addressU = (hex | ch); // so the first - MSB byte - gets over-written
               }
            else                    // count not > 0 (= 0 or-1, this is line crc from file) 
               {                    // check for crc error
               hex = (hex | ch);    // hex is crc at end of line in .hex file
               check_crc -= hex;    // subtract hex, which was added to check_crc above
               check_crc = (~check_crc + 1);  // compute the 2's compliment of check_crc
               if(hex != check_crc) // compare file crc to 2's compliment of check_crc
                  crc_err = 'e';    // set crc error flag if not equal
               }            
         }
      return(ret);
}


void send_preamble()
{
   char cnt = 0;
   unsigned char preamble[4];  // preamble sent before each block

    crc = 0;              // zero the crc to start generating a new one

    preamble[0] = start_addressU; // create preamble sequence with address and size of block
    preamble[1] = (unsigned char)(start_address>>8 & 0x000000ff); // break 16 bit address into 
    preamble[2] = (unsigned char)(start_address & 0x00000ff);     // high and low bytes 
    preamble[3] = 0x80;   // size of block 128 bytes
    for(cnt = 0; cnt < 4; cnt++) // send preamble - this sends addresses to the bootloader
       {
       sndchr = preamble[cnt];
       write(fd,&sndchr,1);
       usleep(100);       // wait for byte to be transmitted (1/115200*10 = 86.8 usec)
       crc += sndchr;     // add byte to crc
       }
}


void send_block()
{
   unsigned char blocksize = 128;
   unsigned char cnt = 0;

    printf(" end of block\n\r"); 
    
    send_preamble();

    while(blocksize-- > 0)  // send 128 byte block from the buffer to bootloader
       {                    
       if(cnt <= buffsize ) // comparing cnt to buffsize handles partial blocks < 128
          {
          sndchr = buffer[cnt++];
          write(fd,&sndchr,1);
          crc += sndchr;
          usleep(100);
          }
       else              // the bootloader is counting bytes received, so
          {              // fill in the rest of a partial block with 0xFF
          sndchr = 0xFF;
          write(fd,&sndchr,1);
          crc += sndchr;
          usleep(100);
          }
       }
    sndchr = (0 - crc);  // when this byte is added (in bootloader), the accumulated crc will = 0
    write(fd,&sndchr,1); // send the last crc byte
    usleep(100);

    read_byte('K');      // wait for 'K' from bootloader
    buffsize = 0;        // set number of bytes in buffer to 0
    usleep(10000);       // wait 10 miliseconds for erase and program cycles to finish
}


void send_byte()          // EEprom and CONFIG are programmed 1 byte at a time by bootloader
{                         // instead of in 128 byte blocks
   unsigned char cnt = 0;
   unsigned char preamble[4]; // preamble sent before each block

     print_terminal(1);       // send the single byte data to print on the terminal
     crc = 0;

     preamble[0] = addressU;
     preamble[1] = (unsigned char)(addr>>8 & 0x000000ff);
     preamble[2] = ((unsigned char)(addr & 0x00000ff) );
     preamble[3] = 0x01;  // single byte
     for(cnt = 0; cnt < 4; cnt++) // send preamble
        {
        sndchr = preamble[cnt];
        write(fd,&sndchr,1);
        usleep(100);      // wait for byte to be transmitted
        crc += sndchr;    // add byte to crc
        }
    sndchr = hex;
    write(fd,&sndchr,1);  // send the data byte
    usleep(100);          // wait for byte to be transmitted

    crc += sndchr;        // add the byte to crc

    sndchr = (0 - crc);   // subtract crc from 0, when this byte is added to accumulated crc, sum will = 0
    write(fd,&sndchr,1);  // send the final crc byte
    usleep(100);          // wait for byte to be transmitted

    read_byte('K');       // wait for 'K' from bootloader
    usleep(10000);        // wait for programming to finish
}


void read_byte(char match)
{   
    // rcvchr = 0;           // normally should be uncommented - commented for debugging
    do
       read(fd, &rcvchr, 1); // keep reading fd until a match is received
    while
       (rcvchr != match);    // check if rcvchr = the intended match character

    printf("%c!\n",rcvchr);  // echo the received byte to terminal
}


void print_terminal(long cntr)
{
 // ------echo the byte put in buffer to terminal------
    if(addressU < 0x02)
       {
       if( (cntr & 0x0000000f) == 0)    // if this is the start of a 16 bit line
          {
          if(cntr == 0)                 // if this is the first byte of program flash
             printf("Program");
          printf("\n\r0%x " ,addressU);
          if((addr & 0x0000ffff) < 0x1000)
             printf("0");
          if((addr & 0x0000ffff) < 0x100)
             printf("0");
          if((addr & 0x0000ffff) < 0x10)
             printf("0");
          printf("%x   ", (addr & 0x0000ffff));
          }
       if(buffsize == 0)
          printf("                                  ");
       if(hex < 0x10)
          printf("0");
       printf("%x ",hex & 0x000ff);
       }
 // ------echo single byte to terminal-----------------
    else
       {
       if(addr == 0)
          printf("\n\r");
       if(addressU == 0x20)
          printf("UserID-");
       if(addressU == 0x31)
          printf("EEProm-");
       if(addressU == 0x30)
          printf("Config-");
       printf("%x ",addressU);
       if((addr & 0x00ffff) < 0x1000)
          printf("0");
       if((addr & 0x00ffff) < 0x100)
          printf("0");
       if((addr & 0x00ffff) < 0x10)
          printf("0");
       printf("%x  ",addr & 0x000ff);
       if(hex < 16)
          printf("0");
       printf("%x-",hex & 0x000ff);
       }
}
